/*
 * Memory quantity calculation utilities.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "memory_utils.h"



/* Return page 'n' for the configured page size. */

offset_t page(unsigned int n)
{
    return PAGE_SIZE * n;
}

/* Return the order of 'size', where 2 ** order yields the size. */

unsigned int page_order(offset_t size)
{
    /* Count zeros from the left, stopping at the first set bit, using the width
       of the size value (in bits, starting with the width in bytes) to
       calculate the position of this bit and thus the order of the value. */

    return sizeof(unsigned long) * 8 - 1 - __builtin_clzl(size);
}

/* Return 'value' rounded up to the nearest 'increment'. */

offset_t round(offset_t value, offset_t increment)
{
    return trunc(value + increment - 1, increment);
}

/* Return 'value' rounded up to the nearest multiple of 'increment'. */

offset_t round_multiple(offset_t value, offset_t increment)
{
    offset_t last = increment;

    while (1)
    {
        if (value < increment)
            return round(value, last);

        last = increment;
        increment *= 2;
    }
}

/* Return 'value' rounded down (or truncated) to the nearest 'increment'. */

offset_t trunc(offset_t value, offset_t increment)
{
    return (value / increment) * increment;
}

/* Return 'value' rounded down (or truncated) to the nearest multiple of
   'increment'. */

offset_t trunc_multiple(offset_t value, offset_t increment)
{
    offset_t last = increment;

    while (1)
    {
        if (value < increment)
            return trunc(value, last);

        last = increment;
        increment *= 2;
    }
}

/* Find the maximum size aligned region within the region from 'start' to (but
   not including) 'end', with the given initial 'increment'. */

offset_t max_multiple(offset_t start, offset_t end, offset_t increment)
{
    /* The largest possible aligned region is derived from the region size. */

    offset_t size = trunc_multiple(end - start, increment);

    /* Apply the alignment to the start. */

    offset_t aligned = round(start, size);

    /* If the region is aligned, return the size. */

    if (aligned == start)
        return size;

    /* If the region is not aligned to the current size, recalculate the aligned
       size. */

    offset_t aligned_size;

    do
    {
        aligned_size = trunc_multiple(end - aligned, increment);
        size /= 2;

        /* Determine whether a smaller alignment could yield a larger aligned
           size. */

        if (aligned_size >= size)
            return aligned_size;

        aligned = round(start, size);
    }
    while (aligned > start);

    return aligned_size;
}

// vim: tabstop=4 expandtab shiftwidth=4
